package solutions.bitbadger.documents.scala.tests.integration

import org.junit.jupiter.api.Assertions.*
import solutions.bitbadger.documents.{Configuration, Dialect, Field, FieldMatch}
import solutions.bitbadger.documents.scala.extensions.*
import solutions.bitbadger.documents.scala.tests.TEST_TABLE

import scala.jdk.CollectionConverters.*

/**
 * Tests for the JSON-returning functions
 *
 * NOTE: PostgreSQL JSONB columns do not preserve the original JSON with which a document was stored. These tests are
 * the most complex within the library, as they have split testing based on the backing data store. The PostgreSQL tests
 * check IDs (and, in the case of ordered queries, which ones occur before which others) vs. the entire JSON string.
 * Meanwhile, SQLite stores JSON as text, and will return exactly the JSON it was given when it was originally written.
 * These tests can ensure the expected round-trip of the entire JSON string.
 */
object JsonFunctions:

  /**
   * PostgreSQL, when returning JSONB as a string, has spaces after commas and colons delineating fields and values.
   * This function will do a crude string replacement to match the target string based on the dialect being tested.
   *
   * @param json The JSON which should be returned
   * @return The actual expected JSON based on the database being tested
   */
  def maybeJsonB(json: String): String =
    Configuration.dialect() match
      case Dialect.SQLITE => json
      case Dialect.POSTGRESQL => json.replace("\":", "\": ").replace(",\"", ", \"")

  /**
   * Create a snippet of JSON to find a document ID
   *
   * @param id The ID of the document
   * @return A connection-aware ID to check for presence and positioning
   */
  private def docId(id: String): String =
    maybeJsonB(s"""{"id":"$id"""")

  def allDefault(db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonAll(TEST_TABLE)
    assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.one), s"Document 'one' not found in JSON ($json)")
        assertTrue(json.contains(JsonDocument.two), s"Document 'two' not found in JSON ($json)")
        assertTrue(json.contains(JsonDocument.three), s"Document 'three' not found in JSON ($json)")
        assertTrue(json.contains(JsonDocument.four), s"Document 'four' not found in JSON ($json)")
        assertTrue(json.contains(JsonDocument.five), s"Document 'five' not found in JSON ($json)")

      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("one")), s"Document 'one' not found in JSON ($json)")
        assertTrue(json.contains(docId("two")), s"Document 'two' not found in JSON ($json)")
        assertTrue(json.contains(docId("three")), s"Document 'three' not found in JSON ($json)")
        assertTrue(json.contains(docId("four")), s"Document 'four' not found in JSON ($json)")
        assertTrue(json.contains(docId("five")), s"Document 'five' not found in JSON ($json)")
    assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def allEmpty(db: ThrowawayDatabase): Unit =
    assertEquals("[]", db.conn.jsonAll(TEST_TABLE), "There should have been no documents returned")

  def byIdString (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonById(TEST_TABLE, "two")
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.two, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("two")), s"An incorrect document was returned ($json)")

  def byIdNumber (db: ThrowawayDatabase): Unit =
    Configuration.idField = "key"
    try {
      db.conn.insert(TEST_TABLE, NumIdDocument(18, "howdy"))
      assertEquals(maybeJsonB("""{"key":18,"text":"howdy"}"""), db.conn.jsonById(TEST_TABLE, 18),
        "The document should have been found by numeric ID")
    } finally {
      Configuration.idField = "id"
    }

  def byIdNotFound (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals("{}", db.conn.jsonById(TEST_TABLE, "x"), "There should have been no document returned")

  def byFieldsMatch (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonByFields(TEST_TABLE,
      Field.any("value", ("blue" :: "purple" :: Nil).asJava) :: Field.exists("sub") :: Nil, Some(FieldMatch.ALL))
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertEquals(s"[${JsonDocument.four}]", json, "The incorrect document was returned")
      case Dialect.POSTGRESQL =>
        assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
        assertTrue(json.contains(docId("four")), s"The incorrect document was returned ($json)")
        assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byFieldsMatchOrdered (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonByFields(TEST_TABLE, Field.equal("value", "purple") :: Nil, None, Field.named("id") :: Nil)
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertEquals(s"[${JsonDocument.five},${JsonDocument.four}]", json, "The documents were not ordered correctly")
      case Dialect.POSTGRESQL =>
        val fiveIdx = json.indexOf(docId("five"))
        val fourIdx = json.indexOf(docId("four"))
        assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
        assertTrue(fiveIdx >= 0, s"Document 'five' not found ($json)")
        assertTrue(fourIdx >= 0, s"Document 'four' not found ($json)")
        assertTrue(fiveIdx < fourIdx, s"Document 'five' should have been before 'four' ($json)")
        assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byFieldsMatchNumIn (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonByFields(TEST_TABLE, Field.any("numValue", (2 :: 4 :: 6 :: 8 :: Nil).asJava) :: Nil)
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertEquals(s"[${JsonDocument.three}]", json, "The incorrect document was returned")
      case Dialect.POSTGRESQL =>
        assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
        assertTrue(json.contains(docId("three")), s"The incorrect document was returned ($json)")
        assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byFieldsNoMatch (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals("[]", db.conn.jsonByFields(TEST_TABLE, Field.greater("numValue", 100) :: Nil),
      "There should have been no documents returned")

  def byFieldsMatchInArray (db: ThrowawayDatabase): Unit =
    ArrayDocument.testDocuments.foreach { doc => db.conn.insert(TEST_TABLE, doc) }
    val json = db.conn.jsonByFields(TEST_TABLE, Field.inArray("values", TEST_TABLE, ("c" :: Nil).asJava) :: Nil)
    assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
    assertTrue(json.contains(docId("first")), s"The 'first' document was not found ($json)")
    assertTrue(json.contains(docId("second")), s"The 'second' document was not found ($json)")
    assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byFieldsNoMatchInArray (db: ThrowawayDatabase): Unit =
    ArrayDocument.testDocuments.foreach { doc => db.conn.insert(TEST_TABLE, doc) }
    assertEquals("[]",
      db.conn.jsonByFields(TEST_TABLE, Field.inArray("values", TEST_TABLE, ("j" :: Nil).asJava) :: Nil),
      "There should have been no documents returned")

  def byContainsMatch (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonByContains(TEST_TABLE, Map.Map1("value", "purple"))
    assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.four), s"Document 'four' not found ($json)")
        assertTrue(json.contains(JsonDocument.five), s"Document 'five' not found ($json)")
      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("four")), s"Document 'four' not found ($json)")
        assertTrue(json.contains(docId("five")), s"Document 'five' not found ($json)")
    assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byContainsMatchOrdered (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonByContains(TEST_TABLE, Map.Map1("sub", Map.Map1("foo", "green")),
      Field.named("value") :: Nil)
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertEquals(s"[${JsonDocument.two},${JsonDocument.four}]", json, "The documents were not ordered correctly")
      case Dialect.POSTGRESQL =>
        val twoIdx = json.indexOf(docId("two"))
        val fourIdx = json.indexOf(docId("four"))
        assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
        assertTrue(twoIdx >= 0, s"Document 'two' not found ($json)")
        assertTrue(fourIdx >= 0, s"Document 'four' not found ($json)")
        assertTrue(twoIdx < fourIdx, s"Document 'two' should have been before 'four' ($json)")
        assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byContainsNoMatch (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals("[]", db.conn.jsonByContains(TEST_TABLE, Map.Map1("value", "indigo")),
      "There should have been no documents returned")

  def byJsonPathMatch (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)")
    assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.four), s"Document 'four' not found ($json)")
        assertTrue(json.contains(JsonDocument.five), s"Document 'five' not found ($json)")
      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("four")), s"Document 'four' not found ($json)")
        assertTrue(json.contains(docId("five")), s"Document 'five' not found ($json)")
    assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byJsonPathMatchOrdered (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)", Field.named("id") :: Nil)
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertEquals(s"[${JsonDocument.five},${JsonDocument.four}]", json, "The documents were not ordered correctly")
      case Dialect.POSTGRESQL =>
        val fiveIdx = json.indexOf(docId("five"))
        val fourIdx = json.indexOf(docId("four"))
        assertTrue(json.startsWith("["), s"JSON should start with '[' ($json)")
        assertTrue(fiveIdx >= 0, s"Document 'five' not found ($json)")
        assertTrue(fourIdx >= 0, s"Document 'four' not found ($json)")
        assertTrue(fiveIdx < fourIdx, s"Document 'five' should have been before 'four' ($json)")
        assertTrue(json.endsWith("]"), s"JSON should end with ']' ($json)")

  def byJsonPathNoMatch (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals("[]", db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 100)"),
      "There should have been no documents returned")

  def firstByFieldsMatchOne (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonFirstByFields(TEST_TABLE, Field.equal("value", "another") :: Nil)
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.two, json, "The incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("two")), s"The incorrect document was returned ($json)")

  def firstByFieldsMatchMany (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonFirstByFields(TEST_TABLE, Field.equal("sub.foo", "green") :: Nil)
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.two) || json.contains(JsonDocument.four),
          s"Expected document 'two' or 'four' ($json)")
      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("two")) || json.contains(docId("four")),
          s"Expected document 'two' or 'four' ($json)")

  def firstByFieldsMatchOrdered (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonFirstByFields(TEST_TABLE, Field.equal("sub.foo", "green") :: Nil, None,
      Field.named("n:numValue DESC") :: Nil)
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.four, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("four")), s"An incorrect document was returned ($json)")

  def firstByFieldsNoMatch (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals("{}", db.conn.jsonFirstByFields(TEST_TABLE, Field.equal("value", "absent") :: Nil),
      "There should have been no document returned")

  def firstByContainsMatchOne (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonFirstByContains(TEST_TABLE, Map.Map1("value", "FIRST!"))
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.one, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("one")), s"An incorrect document was returned ($json)")

  def firstByContainsMatchMany (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonFirstByContains(TEST_TABLE, Map.Map1("value", "purple"))
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
          s"Expected document 'four' or 'five' ($json)")
      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("four")) || json.contains(docId("five")),
          s"Expected document 'four' or 'five' ($json)")

  def firstByContainsMatchOrdered (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonFirstByContains(TEST_TABLE, Map.Map1("value", "purple"),
      Field.named("sub.bar NULLS FIRST") :: Nil)
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.five, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("five")), s"An incorrect document was returned ($json)")

  def firstByContainsNoMatch (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals("{}", db.conn.jsonFirstByContains(TEST_TABLE, Map.Map1("value", "indigo")),
      "There should have been no document returned")

  def firstByJsonPathMatchOne (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ == 10)")
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.two, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("two")), s"An incorrect document was returned ($json)")

  def firstByJsonPathMatchMany (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)")
    Configuration.dialect() match
      case Dialect.SQLITE =>
        assertTrue(json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
          s"Expected document 'four' or 'five' ($json)")
      case Dialect.POSTGRESQL =>
        assertTrue(json.contains(docId("four")) || json.contains(docId("five")),
          s"Expected document 'four' or 'five' ($json)")

  def firstByJsonPathMatchOrdered (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    val json = db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)", Field.named("id DESC") :: Nil)
    Configuration.dialect() match
      case Dialect.SQLITE => assertEquals(JsonDocument.four, json, "An incorrect document was returned")
      case Dialect.POSTGRESQL => assertTrue(json.contains(docId("four")), s"An incorrect document was returned ($json)")

  def firstByJsonPathNoMatch (db: ThrowawayDatabase): Unit =
    JsonDocument.load(db)
    assertEquals("{}", db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 100)"),
      "There should have been no document returned")
